import io
import inspect
import logging
import datetime

from typing import Union
from functools import wraps

import aiohttp

from starlette.responses import StreamingResponse
from logdetective.server.database.models import EndpointType, AnalyzeRequestMetrics
from logdetective.server.remote_log import RemoteLog
from logdetective.server import models
from logdetective.server.compressors import LLMResponseCompressor

LOG = logging.getLogger("logdetective")


async def add_new_metrics(
    api_name: str,
    url: str,
    http_session: aiohttp.ClientSession,
    received_at: datetime.datetime = None,
    compressed_log_content: io.BytesIO = None,
) -> int:
    """Add a new database entry for a received request.

    This will store the time when this function is called,
    the endpoint from where the request was received,
    and the log (in a zip format) for which analysis is requested.
    """
    remote_log = RemoteLog(url, http_session)
    compressed_log_content = compressed_log_content or await remote_log.zip_content()
    return AnalyzeRequestMetrics.create(
        endpoint=EndpointType(api_name),
        compressed_log=compressed_log_content,
        request_received_at=received_at
        if received_at
        else datetime.datetime.now(datetime.timezone.utc),
    )


def update_metrics(
    metrics_id: int,
    response: Union[models.Response, models.StagedResponse, StreamingResponse],
    sent_at: datetime.datetime = None,
) -> None:
    """Update a database metric entry for a received request,
    filling data for the given response.

    This will add to the database entry the time when the response was sent,
    the length of the created response and the certainty for it.
    """
    try:
        compressed_response = LLMResponseCompressor(response).zip_response()
    except AttributeError as e:
        compressed_response = None
        LOG.warning(
            "Given response can not be serialized "
            "and saved in db (probably a StreamingResponse): %s.", e
        )

    response_sent_at = (
        sent_at if sent_at else datetime.datetime.now(datetime.timezone.utc)
    )
    response_length = None
    if hasattr(response, "explanation") and isinstance(
        response.explanation, models.Explanation
    ):
        response_length = len(response.explanation.text)
    response_certainty = (
        response.response_certainty if hasattr(response, "response_certainty") else None
    )
    AnalyzeRequestMetrics.update(
        metrics_id,
        response_sent_at,
        response_length,
        response_certainty,
        compressed_response,
    )


def track_request(name=None):
    """
    Decorator to track requests metrics
    """

    def decorator(f):
        @wraps(f)
        async def async_decorated_function(*args, **kwargs):
            log_url = kwargs["build_log"].url
            metrics_id = await add_new_metrics(
                name if name else f.__name__, log_url, kwargs["http_session"]
            )
            response = await f(*args, **kwargs)
            update_metrics(metrics_id, response)
            return response

        if inspect.iscoroutinefunction(f):
            return async_decorated_function
        raise NotImplementedError("An async coroutine is needed")

    return decorator



import enum
import datetime
from typing import Optional

import backoff

from sqlalchemy import (
    Enum,
    Column,
    BigInteger,
    DateTime,
    String,
    ForeignKey,
    UniqueConstraint,
    desc,
)
from sqlalchemy.orm import relationship
from sqlalchemy.exc import OperationalError
from logdetective.server.database.base import Base, transaction, DB_MAX_RETRIES


class Forge(str, enum.Enum):
    """List of forges managed by logdetective"""

    gitlab_com = "https://gitlab.com"  # pylint: disable=(invalid-name)
    gitlab_cee_redhat_com = "http://gitlab.cee.redhat.com/"  # pylint: disable=(invalid-name)


class GitlabMergeRequestJobs(Base):
    """Store details for the merge request jobs
    which triggered logdetective.
    """

    __tablename__ = "gitlab_merge_request_jobs"

    id = Column(BigInteger, primary_key=True)
    forge = Column(Enum(Forge), nullable=False, index=True, comment="The forge name")
    project_id = Column(
        BigInteger,
        nullable=False,
        index=True,
        comment="The project gitlab id",
    )
    mr_iid = Column(
        BigInteger,
        nullable=False,
        index=False,
        comment="The merge request gitlab iid",
    )
    job_id = Column(
        BigInteger,
        nullable=False,
        index=True,
        comment="The job gitlab id",
    )

    __table_args__ = (
        UniqueConstraint("forge", "job_id", name="uix_forge_job"),
        UniqueConstraint(
            "forge", "project_id", "mr_iid", "job_id", name="uix_mr_project_job"
        ),
    )

    comment = relationship(
        "Comments", back_populates="merge_request_job", uselist=False
    )  # 1 comment for 1 job

    request_metrics = relationship("AnalyzeRequestMetrics", back_populates="mr_job")

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    def create(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
    ) -> int:
        """Create a new merge request job entry,
        returns its PostgreSQL id

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
        """
        with transaction(commit=True) as session:
            mr = cls()
            mr.forge = forge
            mr.project_id = project_id
            mr.mr_iid = mr_iid
            mr.job_id = job_id
            session.add(mr)
            session.flush()
            return mr.id

    @classmethod
    def get_by_id(
        cls,
        id_: int,
    ) -> Optional["GitlabMergeRequestJobs"]:
        """Search for a given PostgreSQL id"""
        with transaction(commit=False) as session:
            mr = session.query(cls).filter_by(id=id_).first()
            return mr

    @classmethod
    def get_by_details(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
    ) -> Optional["GitlabMergeRequestJobs"]:
        """Search for a detailed merge request.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
        """
        with transaction(commit=False) as session:
            mr = (
                session.query(cls)
                .filter_by(
                    forge=forge, project_id=project_id, mr_iid=mr_iid, job_id=job_id
                )
                .first()
            )
            return mr

    @classmethod
    def get_or_create(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
    ) -> Optional["GitlabMergeRequestJobs"]:
        """Search for a detailed merge request
        or create a new one if not found.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
        """
        mr = GitlabMergeRequestJobs.get_by_details(forge, project_id, mr_iid, job_id)
        if mr is None:
            id_ = GitlabMergeRequestJobs.create(forge, project_id, mr_iid, job_id)
            mr = GitlabMergeRequestJobs.get_by_id(id_)
        return mr


class Comments(Base):
    """Store details for comments
    created by logdetective for a merge request job."""

    __tablename__ = "comments"

    id = Column(BigInteger, primary_key=True)
    merge_request_job_id = Column(
        BigInteger,
        ForeignKey("gitlab_merge_request_jobs.id"),
        nullable=False,
        unique=True,  # 1 comment for 1 job
        index=True,
        comment="The associated merge request job (db) id",
    )
    forge = Column(
        Enum(Forge),
        nullable=False,
        index=True,
        comment="The forge name"
    )
    comment_id = Column(
        String(50),  # e.g. 'd5a3ff139356ce33e37e73add446f16869741b50'
        nullable=False,
        index=True,
        comment="The comment gitlab id",
    )
    created_at = Column(
        DateTime, nullable=False, comment="Timestamp when the comment was created"
    )

    __table_args__ = (
        UniqueConstraint("forge", "comment_id", name="uix_forge_comment_id"),
    )

    merge_request_job = relationship("GitlabMergeRequestJobs", back_populates="comment")
    reactions = relationship("Reactions", back_populates="comment")

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    def create(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
        metrics: Optional["AnalyzeRequestMetrics"] = None,  # noqa: F821
    ) -> int:
        """Create a new comment id entry,
        returns its PostgreSQL id.
        A Gitlab comment id is unique within a
        gitlab instance.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
        """
        with transaction(commit=True) as session:
            mr_job = GitlabMergeRequestJobs.get_or_create(
                forge, project_id, mr_iid, job_id
            )

            if metrics:
                metrics.mr_job = mr_job
                session.add(metrics)

            comment = cls()
            comment.forge = forge
            comment.comment_id = comment_id
            comment.created_at = datetime.datetime.now(datetime.timezone.utc)
            comment.merge_request_job_id = mr_job.id
            session.add(comment)
            session.flush()
            return comment.id

    @classmethod
    def get_by_id(
        cls,
        id_: int,
    ) -> Optional["Comments"]:
        """Search for a given PostgreSQL id"""
        with transaction(commit=False) as session:
            comment = session.query(cls).filter_by(id=id_).first()
            return comment

    @classmethod
    def get_by_gitlab_id(
        cls,
        forge: Forge,
        comment_id: str,
    ) -> Optional["Comments"]:
        """Search for a detailed comment
        by its unique forge comment id.

        Args:
          forge: forge name
          comment_id: forge comment id
        """
        with transaction(commit=False) as session:
            comment = (
                session.query(cls)
                .join(
                    GitlabMergeRequestJobs,
                    cls.merge_request_job_id == GitlabMergeRequestJobs.id,
                )
                .filter(
                    GitlabMergeRequestJobs.forge == forge,
                    cls.comment_id == comment_id,
                )
                .first()
            )

            return comment

    @classmethod
    def get_latest_comment(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
    ) -> Optional["Comments"]:
        """Search for the latest comment in the merge request.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
        """
        with transaction(commit=False) as session:
            comment = (
                session.query(cls)
                .join(
                    GitlabMergeRequestJobs,
                    cls.merge_request_job_id == GitlabMergeRequestJobs.id,
                )
                .filter(
                    GitlabMergeRequestJobs.forge == forge,
                    GitlabMergeRequestJobs.project_id == project_id,
                    GitlabMergeRequestJobs.mr_iid == mr_iid,
                )
                .order_by(desc(cls.created_at))
                .first()
            )

            return comment

    @classmethod
    def get_mr_comments(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
    ) -> Optional["Comments"]:
        """Search for all merge request comments.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
        """
        with transaction(commit=False) as session:
            comments = (
                session.query(cls)
                .join(
                    GitlabMergeRequestJobs,
                    cls.merge_request_job_id == GitlabMergeRequestJobs.id,
                )
                .filter(
                    GitlabMergeRequestJobs.forge == forge,
                    GitlabMergeRequestJobs.project_id == project_id,
                    GitlabMergeRequestJobs.mr_iid == mr_iid,
                )
                .order_by(desc(cls.created_at))
                .all()
            )

            return comments

    @classmethod
    def get_or_create(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
    ) -> Optional["Comments"]:
        """Search for a detailed comment
        or create a new one if not found.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
        """
        comment = Comments.get_by_gitlab_id(forge, comment_id)
        if comment is None:
            id_ = Comments.create(forge, project_id, mr_iid, job_id, comment_id)
            comment = GitlabMergeRequestJobs.get_by_id(id_)
        return comment


class Reactions(Base):
    """Store and count reactions received for
    logdetective comments"""

    __tablename__ = "reactions"

    id = Column(BigInteger, primary_key=True)
    comment_id = Column(
        BigInteger,
        ForeignKey("comments.id"),
        nullable=False,
        index=True,
        comment="The associated comment (db) id",
    )
    reaction_type = Column(
        String(127),  # e.g. 'thumbs-up'
        nullable=False,
        comment="The type of reaction",
    )
    count = Column(
        BigInteger,
        nullable=False,
        comment="The number of reactions, of this type, given in the comment",
    )

    __table_args__ = (
        UniqueConstraint("comment_id", "reaction_type", name="uix_comment_reaction"),
    )

    comment = relationship("Comments", back_populates="reactions")

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    def create_or_update(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
        reaction_type: str,
        count: int,
    ) -> int:
        """Create or update the given reaction, and its associated count,
        for a specified comment.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
          reaction_type: a str, ex. thumb_up
          count: number of reactions, of this type, given in the comment
        """
        comment = Comments.get_or_create(forge, project_id, mr_iid, job_id, comment_id)
        reaction = cls.get_reaction_by_type(
            forge, project_id, mr_iid, job_id, comment_id, reaction_type
        )
        with transaction(commit=True) as session:
            if reaction:
                reaction.count = count  # just update
            else:
                reaction = cls()
                reaction.comment_id = comment.id
                reaction.reaction_type = reaction_type
                reaction.count = count
            session.add(reaction)
            session.flush()
            return reaction.id

    @classmethod
    def get_all_reactions(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
    ) -> int:
        """Get all reactions for a comment

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
        """
        with transaction(commit=False) as session:
            reactions = (
                session.query(cls)
                .join(Comments, cls.comment_id == Comments.id)
                .join(
                    GitlabMergeRequestJobs,
                    Comments.merge_request_job_id == GitlabMergeRequestJobs.id,
                )
                .filter(
                    Comments.comment_id == comment_id,
                    GitlabMergeRequestJobs.forge == forge,
                    GitlabMergeRequestJobs.project_id == project_id,
                    GitlabMergeRequestJobs.mr_iid == mr_iid,
                    GitlabMergeRequestJobs.job_id == job_id,
                )
                .all()
            )

            return reactions

    @classmethod
    def get_reaction_by_type(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
        reaction_type: str,
    ) -> int:
        """Get reaction, of a given type,
        for a comment

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
          reaction_type: str like "thumb-up"
        """
        with transaction(commit=False) as session:
            reaction = (
                session.query(cls)
                .join(Comments, cls.comment_id == Comments.id)
                .join(
                    GitlabMergeRequestJobs,
                    Comments.merge_request_job_id == GitlabMergeRequestJobs.id,
                )
                .filter(
                    Comments.comment_id == comment_id,
                    GitlabMergeRequestJobs.forge == forge,
                    GitlabMergeRequestJobs.project_id == project_id,
                    GitlabMergeRequestJobs.mr_iid == mr_iid,
                    GitlabMergeRequestJobs.job_id == job_id,
                    Reactions.reaction_type == reaction_type,
                )
                .first()
            )

            return reaction



from typing import Optional
from pydantic import BaseModel

from logdetective.constants import (
    PROMPT_TEMPLATE,
    PROMPT_TEMPLATE_STAGED,
    SUMMARIZATION_PROMPT_TEMPLATE,
    SNIPPET_PROMPT_TEMPLATE,
)


class PromptConfig(BaseModel):
    """Configuration for basic log detective prompts."""

    prompt_template: str = PROMPT_TEMPLATE
    summarization_prompt_template: str = SUMMARIZATION_PROMPT_TEMPLATE
    snippet_prompt_template: str = SNIPPET_PROMPT_TEMPLATE
    prompt_template_staged: str = PROMPT_TEMPLATE_STAGED

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return
        self.prompt_template = data.get("prompt_template", PROMPT_TEMPLATE)
        self.summarization_prompt_template = data.get(
            "summarization_prompt_template", SUMMARIZATION_PROMPT_TEMPLATE
        )
        self.snippet_prompt_template = data.get(
            "snippet_prompt_template", SNIPPET_PROMPT_TEMPLATE
        )
        self.prompt_template_staged = data.get(
            "prompt_template_staged", PROMPT_TEMPLATE_STAGED
        )



import asyncio
import json
import os
import re
import zipfile
from enum import Enum
from contextlib import asynccontextmanager
from pathlib import Path, PurePath
from tempfile import TemporaryFile
from typing import List, Annotated, Tuple, Dict, Any, Union
from io import BytesIO

import backoff
import matplotlib
import matplotlib.pyplot
from aiohttp import StreamReader
from fastapi import FastAPI, HTTPException, BackgroundTasks, Depends, Header, Request

from fastapi.responses import StreamingResponse
from fastapi.responses import Response as BasicResponse
import gitlab
import gitlab.v4
import gitlab.v4.objects
import jinja2
import aiohttp
import sqlalchemy
import sentry_sdk

import logdetective.server.database.base

from logdetective.extractors import DrainExtractor
from logdetective.utils import (
    compute_certainty,
    format_snippets,
    load_prompts,
)
from logdetective.server.utils import (
    load_server_config,
    get_log,
    format_analyzed_snippets,
)
from logdetective.server.metric import track_request, add_new_metrics, update_metrics
from logdetective.server.models import (
    BuildLog,
    JobHook,
    Response,
    StagedResponse,
    Explanation,
    AnalyzedSnippet,
    TimePeriod,
)
from logdetective.server import plot as plot_engine
from logdetective.server.remote_log import RemoteLog
from logdetective.server.database.models import (
    Comments,
    EndpointType,
    Forge,
)
from logdetective.server.database.models import AnalyzeRequestMetrics

LLM_CPP_SERVER_TIMEOUT = os.environ.get("LLAMA_CPP_SERVER_TIMEOUT", 600)
LOG_SOURCE_REQUEST_TIMEOUT = os.environ.get("LOG_SOURCE_REQUEST_TIMEOUT", 60)
API_TOKEN = os.environ.get("LOGDETECTIVE_TOKEN", None)
SERVER_CONFIG_PATH = os.environ.get("LOGDETECTIVE_SERVER_CONF", None)
SERVER_PROMPT_PATH = os.environ.get("LOGDETECTIVE_PROMPTS", None)

SERVER_CONFIG = load_server_config(SERVER_CONFIG_PATH)
PROMPT_CONFIG = load_prompts(SERVER_PROMPT_PATH)

MR_REGEX = re.compile(r"refs/merge-requests/(\d+)/.*$")
FAILURE_LOG_REGEX = re.compile(r"(\w*\.log)")

LOG = get_log(SERVER_CONFIG)

if sentry_dsn := SERVER_CONFIG.general.sentry_dsn:
    sentry_sdk.init(dsn=str(sentry_dsn), traces_sample_rate=1.0)


@asynccontextmanager
async def lifespan(fapp: FastAPI):
    """
    Establish one HTTP session
    """
    fapp.http = aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(
            total=int(LOG_SOURCE_REQUEST_TIMEOUT), connect=3.07
        )
    )

    # Ensure that the database is initialized.
    logdetective.server.database.base.init()

    yield
    await fapp.http.close()


async def get_http_session(request: Request) -> aiohttp.ClientSession:
    """
    Return the single aiohttp ClientSession for this app
    """
    return request.app.http


def requires_token_when_set(authentication: Annotated[str | None, Header()] = None):
    """
    FastAPI Depend function that expects a header named Authentication

    If LOGDETECTIVE_TOKEN env var is set, validate the client-supplied token
    otherwise ignore it
    """
    if not API_TOKEN:
        LOG.info("LOGDETECTIVE_TOKEN env var not set, authentication disabled")
        # no token required, means local dev environment
        return
    token = None
    if authentication:
        try:
            token = authentication.split(" ", 1)[1]
        except (ValueError, IndexError):
            LOG.warning(
                "Authentication header has invalid structure (%s), it should be 'Bearer TOKEN'",
                authentication,
            )
            # eat the exception and raise 401 below
            token = None
        if token == API_TOKEN:
            return
    LOG.info(
        "LOGDETECTIVE_TOKEN env var is set (%s), clien token = %s", API_TOKEN, token
    )
    raise HTTPException(status_code=401, detail=f"Token {token} not valid.")


app = FastAPI(dependencies=[Depends(requires_token_when_set)], lifespan=lifespan)
app.gitlab_conn = gitlab.Gitlab(
    url=SERVER_CONFIG.gitlab.url, private_token=SERVER_CONFIG.gitlab.api_token
)


def mine_logs(log: str) -> List[Tuple[int, str]]:
    """Extract snippets from log text"""
    extractor = DrainExtractor(
        verbose=True, context=True, max_clusters=SERVER_CONFIG.extractor.max_clusters
    )

    LOG.info("Getting summary")
    log_summary = extractor(log)

    ratio = len(log_summary) / len(log.split("\n"))
    LOG.debug("Log summary: \n %s", log_summary)
    LOG.info("Compression ratio: %s", ratio)

    return log_summary


async def submit_to_llm_endpoint(
    http: aiohttp.ClientSession,
    url: str,
    data: Dict[str, Any],
    headers: Dict[str, str],
    stream: bool,
) -> Any:
    """Send request to selected API endpoint. Verifying successful request unless
    the using the stream response.

    url:
    data:
    headers:
    stream:
    """
    LOG.debug("async request %s headers=%s data=%s", url, headers, data)
    response = await http.post(
        url,
        headers=headers,
        # we need to use the `json=` parameter here and let aiohttp
        # handle the json-encoding
        json=data,
        timeout=int(LLM_CPP_SERVER_TIMEOUT),
        # Docs says chunked takes int, but:
        #   DeprecationWarning: Chunk size is deprecated #1615
        # So let's make sure we either put True or None here
        chunked=True if stream else None,
        raise_for_status=True,
    )
    if stream:
        return response
    try:
        return json.loads(await response.text())
    except UnicodeDecodeError as ex:
        LOG.error("Error encountered while parsing llama server response: %s", ex)
        raise HTTPException(
            status_code=400,
            detail=f"Couldn't parse the response.\nError: {ex}\nData: {response.text}",
        ) from ex


def should_we_giveup(exc: aiohttp.ClientResponseError) -> bool:
    """
    From backoff's docs:

    > a function which accepts the exception and returns
    > a truthy value if the exception should not be retried
    """
    LOG.info("Should we give up on retrying error %s", exc)
    return exc.status < 500


def we_give_up(details: backoff._typing.Details):
    """
    retries didn't work (or we got a different exc)
    we give up and raise proper 500 for our API endpoint
    """
    LOG.error("Inference error: %s", details["args"])
    raise HTTPException(500, "Request to the inference API failed")


@backoff.on_exception(
    backoff.expo,
    aiohttp.ClientResponseError,
    max_tries=3,
    giveup=should_we_giveup,
    raise_on_giveup=False,
    on_giveup=we_give_up,
)
async def submit_text(  # pylint: disable=R0913,R0917
    http: aiohttp.ClientSession,
    text: str,
    max_tokens: int = -1,
    log_probs: int = 1,
    stream: bool = False,
    model: str = "default-model",
) -> Explanation:
    """Submit prompt to LLM using a selected endpoint.
    max_tokens: number of tokens to be produces, 0 indicates run until encountering EOS
    log_probs: number of token choices to produce log probs for
    """
    LOG.info("Analyzing the text")

    headers = {"Content-Type": "application/json"}

    if SERVER_CONFIG.inference.api_token:
        headers["Authorization"] = f"Bearer {SERVER_CONFIG.inference.api_token}"

    if SERVER_CONFIG.inference.api_endpoint == "/chat/completions":
        return await submit_text_chat_completions(
            http, text, headers, max_tokens, log_probs > 0, stream, model
        )
    return await submit_text_completions(
        http, text, headers, max_tokens, log_probs, stream, model
    )


async def submit_text_completions(  # pylint: disable=R0913,R0917
    http: aiohttp.ClientSession,
    text: str,
    headers: dict,
    max_tokens: int = -1,
    log_probs: int = 1,
    stream: bool = False,
    model: str = "default-model",
) -> Explanation:
    """Submit prompt to OpenAI API completions endpoint.
    max_tokens: number of tokens to be produces, 0 indicates run until encountering EOS
    log_probs: number of token choices to produce log probs for
    """
    LOG.info("Submitting to /v1/completions endpoint")
    data = {
        "prompt": text,
        "max_tokens": max_tokens,
        "logprobs": log_probs,
        "stream": stream,
        "model": model,
        "temperature": SERVER_CONFIG.inference.temperature,
    }

    response = await submit_to_llm_endpoint(
        http,
        f"{SERVER_CONFIG.inference.url}/v1/completions",
        data,
        headers,
        stream,
    )

    return Explanation(
        text=response["choices"][0]["text"], logprobs=response["choices"][0]["logprobs"]
    )


async def submit_text_chat_completions(  # pylint: disable=R0913,R0917
    http: aiohttp.ClientSession,
    text: str,
    headers: dict,
    max_tokens: int = -1,
    log_probs: int = 1,
    stream: bool = False,
    model: str = "default-model",
) -> Union[Explanation, StreamReader]:
    """Submit prompt to OpenAI API /chat/completions endpoint.
    max_tokens: number of tokens to be produces, 0 indicates run until encountering EOS
    log_probs: number of token choices to produce log probs for
    """
    LOG.info("Submitting to /v1/chat/completions endpoint")

    data = {
        "messages": [
            {
                "role": "user",
                "content": text,
            }
        ],
        "max_tokens": max_tokens,
        "logprobs": log_probs,
        "stream": stream,
        "model": model,
        "temperature": SERVER_CONFIG.inference.temperature,
    }

    response = await submit_to_llm_endpoint(
        http,
        f"{SERVER_CONFIG.inference.url}/v1/chat/completions",
        data,
        headers,
        stream,
    )

    if stream:
        return response
    return Explanation(
        text=response["choices"][0]["message"]["content"],
        logprobs=response["choices"][0]["logprobs"]["content"],
    )


@app.post("/analyze", response_model=Response)
@track_request()
async def analyze_log(
    build_log: BuildLog, http_session: aiohttp.ClientSession = Depends(get_http_session)
):
    """Provide endpoint for log file submission and analysis.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    remote_log = RemoteLog(build_log.url, http_session)
    log_text = await remote_log.process_url()
    log_summary = mine_logs(log_text)
    log_summary = format_snippets(log_summary)
    response = await submit_text(
        http_session,
        PROMPT_CONFIG.prompt_template.format(log_summary),
        model=SERVER_CONFIG.inference.model,
        max_tokens=SERVER_CONFIG.inference.max_tokens,
    )
    certainty = 0

    if response.logprobs is not None:
        try:
            certainty = compute_certainty(response.logprobs)
        except ValueError as ex:
            LOG.error("Error encountered while computing certainty: %s", ex)
            raise HTTPException(
                status_code=400,
                detail=f"Couldn't compute certainty with data:\n{response.logprobs}",
            ) from ex

    return Response(explanation=response, response_certainty=certainty)


@track_request()
@app.post("/analyze/staged", response_model=StagedResponse)
async def analyze_log_staged(
    build_log: BuildLog, http_session: aiohttp.ClientSession = Depends(get_http_session)
):
    """Provide endpoint for log file submission and analysis.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    remote_log = RemoteLog(build_log.url, http_session)
    log_text = await remote_log.process_url()

    return await perform_staged_analysis(http_session, log_text=log_text)


async def perform_staged_analysis(
    http: aiohttp.ClientSession, log_text: str
) -> StagedResponse:
    """Submit the log file snippets to the LLM and retrieve their results"""
    log_summary = mine_logs(log_text)

    # Process snippets asynchronously
    analyzed_snippets = await asyncio.gather(
        *[
            submit_text(
                http,
                PROMPT_CONFIG.snippet_prompt_template.format(s),
                model=SERVER_CONFIG.inference.model,
                max_tokens=SERVER_CONFIG.inference.max_tokens,
            )
            for s in log_summary
        ]
    )

    analyzed_snippets = [
        AnalyzedSnippet(line_number=e[0][0], text=e[0][1], explanation=e[1])
        for e in zip(log_summary, analyzed_snippets)
    ]
    final_prompt = PROMPT_CONFIG.prompt_template_staged.format(
        format_analyzed_snippets(analyzed_snippets)
    )

    final_analysis = await submit_text(
        http,
        final_prompt,
        model=SERVER_CONFIG.inference.model,
        max_tokens=SERVER_CONFIG.inference.max_tokens,
    )

    certainty = 0

    if final_analysis.logprobs:
        try:
            certainty = compute_certainty(final_analysis.logprobs)
        except ValueError as ex:
            LOG.error("Error encountered while computing certainty: %s", ex)
            raise HTTPException(
                status_code=400,
                detail=f"Couldn't compute certainty with data:\n"
                f"{final_analysis.logprobs}",
            ) from ex

    return StagedResponse(
        explanation=final_analysis,
        snippets=analyzed_snippets,
        response_certainty=certainty,
    )


@app.post("/analyze/stream", response_class=StreamingResponse)
@track_request()
async def analyze_log_stream(
    build_log: BuildLog, http_session: aiohttp.ClientSession = Depends(get_http_session)
):
    """Stream response endpoint for Logdetective.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    remote_log = RemoteLog(build_log.url, http_session)
    log_text = await remote_log.process_url()
    log_summary = mine_logs(log_text)
    log_summary = format_snippets(log_summary)
    headers = {"Content-Type": "application/json"}

    if SERVER_CONFIG.inference.api_token:
        headers["Authorization"] = f"Bearer {SERVER_CONFIG.inference.api_token}"

    try:
        stream = await submit_text_chat_completions(
            http_session,
            PROMPT_CONFIG.prompt_template.format(log_summary),
            stream=True,
            headers=headers,
            model=SERVER_CONFIG.inference.model,
            max_tokens=SERVER_CONFIG.inference.max_tokens,
        )
    except aiohttp.ClientResponseError as ex:
        raise HTTPException(
            status_code=400,
            detail="HTTP Error while getting response from inference server "
            f"[{ex.status}] {ex.message}",
        ) from ex

    # we need to figure out a better response here, this is how it looks rn:
    # b'data: {"choices":[{"finish_reason":"stop","index":0,"delta":{}}],
    #   "created":1744818071,"id":"chatcmpl-c9geTxNcQO7M9wR...
    return StreamingResponse(stream)


@app.post("/webhook/gitlab/job_events")
async def receive_gitlab_job_event_webhook(
    x_gitlab_instance: Annotated[str | None, Header()],
    job_hook: JobHook,
    background_tasks: BackgroundTasks,
    http: aiohttp.ClientSession = Depends(get_http_session),
):
    """Webhook endpoint for receiving job_events notifications from GitLab
    https://docs.gitlab.com/user/project/integrations/webhook_events/#job-events
    lists the full specification for the messages sent for job events."""

    try:
        forge = Forge(x_gitlab_instance)
    except ValueError:
        LOG.critical("%s is not a recognized forge. Ignoring.", x_gitlab_instance)
        return BasicResponse(status_code=400)

    # Handle the message in the background so we can return 200 immediately
    background_tasks.add_task(process_gitlab_job_event, http, forge, job_hook)

    # No return value or body is required for a webhook.
    # 204: No Content
    return BasicResponse(status_code=204)


async def process_gitlab_job_event(
    http: aiohttp.ClientSession,
    forge: Forge,
    job_hook: JobHook,
):
    """Handle a received job_event webhook from GitLab"""
    LOG.debug("Received webhook message from %s:\n%s", forge.value, job_hook)

    # Look up the project this job belongs to
    project = await asyncio.to_thread(app.gitlab_conn.projects.get, job_hook.project_id)
    LOG.info("Processing failed job for %s", project.name)

    # Retrieve data about the job from the GitLab API
    job = await asyncio.to_thread(project.jobs.get, job_hook.build_id)

    # For easy retrieval later, we'll add project_name and project_url to the
    # job object
    job.project_name = project.name
    job.project_url = project.web_url

    # Retrieve the pipeline that started this job
    pipeline = await asyncio.to_thread(project.pipelines.get, job_hook.pipeline_id)

    # Verify this is a merge request
    if pipeline.source != "merge_request_event":
        LOG.info("Not a merge request pipeline. Ignoring.")
        return

    # Extract the merge-request ID from the job
    match = MR_REGEX.search(pipeline.ref)
    if not match:
        LOG.error(
            "Pipeline source is merge_request_event but no merge request ID was provided."
        )
        return
    merge_request_iid = int(match.group(1))

    LOG.debug("Retrieving log artifacts")
    # Retrieve the build logs from the merge request artifacts and preprocess them
    try:
        log_url, preprocessed_log = await retrieve_and_preprocess_koji_logs(http, job)
    except LogsTooLargeError:
        LOG.error("Could not retrieve logs. Too large.")
        raise

    # Submit log to Log Detective and await the results.
    log_text = preprocessed_log.read().decode(encoding="utf-8")
    metrics_id = await add_new_metrics(
        api_name=EndpointType.ANALYZE_GITLAB_JOB,
        url=log_url,
        http_session=http,
        compressed_log_content=RemoteLog.zip_text(log_text),
    )
    staged_response = await perform_staged_analysis(http, log_text=log_text)
    update_metrics(metrics_id, staged_response)
    preprocessed_log.close()

    # check if this project is on the opt-in list for posting comments.
    if project.name not in SERVER_CONFIG.general.packages:
        LOG.info("Not publishing comment for unrecognized package %s", project.name)
        return

    # Add the Log Detective response as a comment to the merge request
    await comment_on_mr(
        forge,
        project,
        merge_request_iid,
        job,
        log_url,
        staged_response,
        metrics_id,
    )

    return staged_response


class LogsTooLargeError(RuntimeError):
    """The log archive exceeds the configured maximum size"""


async def retrieve_and_preprocess_koji_logs(
    http: aiohttp.ClientSession, job: gitlab.v4.objects.ProjectJob
):  # pylint: disable=too-many-branches
    """Download logs from the merge request artifacts

    This function will retrieve the build logs and do some minimal
    preprocessing to determine which log is relevant for analysis.

    returns: The URL pointing to the selected log file and an open, file-like
    object containing the log contents to be sent for processing by Log
    Detective. The calling function is responsible for closing this object."""

    # Make sure the file isn't too large to process.
    if not await check_artifacts_file_size(http, job):
        raise LogsTooLargeError(
            f"Oversized logs for job {job.id} in project {job.project_id}"
        )

    # Create a temporary file to store the downloaded log zipfile.
    # This will be automatically deleted when the last reference into it
    # (returned by this function) is closed.
    tempfile = TemporaryFile(mode="w+b")
    await asyncio.to_thread(job.artifacts, streamed=True, action=tempfile.write)
    tempfile.seek(0)

    failed_arches = {}
    artifacts_zip = zipfile.ZipFile(tempfile, mode="r")  # pylint: disable=consider-using-with
    for zipinfo in artifacts_zip.infolist():
        if zipinfo.filename.endswith("task_failed.log"):
            # The koji logs store this file in two places: 1) in the
            # directory with the failed architecture and 2) in the parent
            # directory. Most of the time, we want to ignore the one in the
            # parent directory, since the rest of the information is in the
            # specific task directory. However, there are some situations
            # where non-build failures (such as "Target build already exists")
            # may be presented only at the top level.
            # The paths look like `kojilogs/noarch-XXXXXX/task_failed.log`
            # or `kojilogs/noarch-XXXXXX/x86_64-XXXXXX/task_failed.log`
            path = PurePath(zipinfo.filename)
            if len(path.parts) <= 3:
                failed_arches["toplevel"] = path
                continue

            # Extract the architecture from the immediate parent path
            architecture = path.parent.parts[-1].split("-")[0]

            # Open this file and read which log failed.
            # The string in this log has the format
            # `see <log> for more information`.
            # Note: it may sometimes say
            # `see build.log or root.log for more information`, but in
            # that situation, we only want to handle build.log (for now),
            # which means accepting only the first match for the regular
            # expression.
            with artifacts_zip.open(zipinfo.filename) as task_failed_log:
                contents = task_failed_log.read().decode("utf-8")
                match = FAILURE_LOG_REGEX.search(contents)
                if not match:
                    LOG.error(
                        "task_failed.log does not indicate which log contains the failure."
                    )
                    raise SyntaxError(
                        "task_failed.log does not indicate which log contains the failure."
                    )
                failure_log_name = match.group(1)

            failed_arches[architecture] = PurePath(path.parent, failure_log_name)

    if not failed_arches:
        # No failed task found in the sub-tasks.
        raise FileNotFoundError("Could not detect failed architecture.")

    # We only want to handle one arch, so we'll check them in order of
    # "most to least likely for the maintainer to have access to hardware"
    # This means: x86_64 > aarch64 > riscv > ppc64le > s390x
    if "x86_64" in failed_arches:
        failed_arch = "x86_64"
    elif "aarch64" in failed_arches:
        failed_arch = "aarch64"
    elif "riscv" in failed_arches:
        failed_arch = "riscv"
    elif "ppc64le" in failed_arches:
        failed_arch = "ppc64le"
    elif "s390x" in failed_arches:
        failed_arch = "s390x"
    elif "noarch" in failed_arches:
        # May have failed during BuildSRPMFromSCM phase
        failed_arch = "noarch"
    elif "toplevel" in failed_arches:
        # Probably a Koji-specific error, not a build error
        failed_arch = "toplevel"
    else:
        # We have one or more architectures that we don't know about? Just
        # pick the first alphabetically.
        failed_arch = sorted(list(failed_arches.keys()))[0]

    LOG.debug("Failed architecture: %s", failed_arch)

    log_path = failed_arches[failed_arch].as_posix()

    log_url = f"{SERVER_CONFIG.gitlab.api_url}/projects/{job.project_id}/jobs/{job.id}/artifacts/{log_path}"  # pylint: disable=line-too-long
    LOG.debug("Returning contents of %s", log_url)

    # Return the log as a file-like object with .read() function
    return log_url, artifacts_zip.open(log_path)


async def check_artifacts_file_size(
    http: aiohttp.ClientSession,
    job: gitlab.v4.objects.ProjectJob,
):
    """Method to determine if the artifacts are too large to process"""
    # First, make sure that the artifacts are of a reasonable size. The
    # zipped artifact collection will be stored in memory below. The
    # python-gitlab library doesn't expose a way to check this value directly,
    # so we need to interact with directly with the headers.
    artifacts_url = f"{SERVER_CONFIG.gitlab.api_url}/projects/{job.project_id}/jobs/{job.id}/artifacts"  # pylint: disable=line-too-long
    LOG.debug("checking artifact URL %s", artifacts_url)
    try:
        head_response = await http.head(
            artifacts_url,
            allow_redirects=True,
            headers={"Authorization": f"Bearer {SERVER_CONFIG.gitlab.api_token}"},
            timeout=5,
            raise_for_status=True,
        )
    except aiohttp.ClientResponseError as ex:
        raise HTTPException(
            status_code=400,
            detail=f"Unable to check artifact URL: [{ex.status}] {ex.message}",
        ) from ex
    content_length = int(head_response.headers.get("content-length"))
    LOG.debug(
        "URL: %s, content-length: %d, max length: %d",
        artifacts_url,
        content_length,
        SERVER_CONFIG.gitlab.max_artifact_size,
    )
    return content_length <= SERVER_CONFIG.gitlab.max_artifact_size


async def comment_on_mr(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
    forge: Forge,
    project: gitlab.v4.objects.Project,
    merge_request_iid: int,
    job: gitlab.v4.objects.ProjectJob,
    log_url: str,
    response: StagedResponse,
    metrics_id: int,
):
    """Add the Log Detective response as a comment to the merge request"""
    LOG.debug(
        "Primary Explanation for %s MR %d: %s",
        project.name,
        merge_request_iid,
        response.explanation.text,
    )

    # First, we'll see if there's an existing comment on this Merge Request
    # and wrap it in <details></details> to reduce noise.
    await suppress_latest_comment(forge, project, merge_request_iid)

    # Get the formatted short comment.
    short_comment = await generate_mr_comment(job, log_url, response, full=False)

    # Look up the merge request
    merge_request = await asyncio.to_thread(
        project.mergerequests.get, merge_request_iid
    )

    # Submit a new comment to the Merge Request using the Gitlab API
    discussion = await asyncio.to_thread(
        merge_request.discussions.create, {"body": short_comment}
    )

    # Get the ID of the first note
    note_id = discussion.attributes["notes"][0]["id"]
    note = discussion.notes.get(note_id)

    # Update the comment with the full details
    # We do this in a second step so we don't bombard the user's email
    # notifications with a massive message. Gitlab doesn't send email for
    # comment edits.
    full_comment = await generate_mr_comment(job, log_url, response, full=True)
    note.body = full_comment

    # Pause for five seconds before sending the snippet data, otherwise
    # Gitlab may bundle the edited message together with the creation
    # message in email.
    await asyncio.sleep(5)
    await asyncio.to_thread(note.save)

    # Save the new comment to the database
    try:
        metrics = AnalyzeRequestMetrics.get_metric_by_id(metrics_id)
        Comments.create(
            forge,
            project.id,
            merge_request_iid,
            job.id,
            discussion.id,
            metrics,
        )
    except sqlalchemy.exc.IntegrityError:
        # We most likely attempted to save a new comment for the same
        # build job. This is somewhat common during development when we're
        # submitting requests manually. It shouldn't really happen in
        # production.
        if not SERVER_CONFIG.general.devmode:
            raise


async def suppress_latest_comment(
    gitlab_instance: str,
    project: gitlab.v4.objects.Project,
    merge_request_iid: int,
) -> None:
    """Look up the latest comment on this Merge Request, if any, and wrap it
    in a <details></details> block with a comment indicating that it has been
    superseded by a new push."""

    # Ask the database for the last known comment for this MR
    previous_comment = Comments.get_latest_comment(
        gitlab_instance, project.id, merge_request_iid
    )

    if previous_comment is None:
        # No existing comment, so nothing to do.
        return

    # Retrieve its content from the Gitlab API

    # Look up the merge request
    merge_request = await asyncio.to_thread(
        project.mergerequests.get, merge_request_iid
    )

    # Find the discussion matching the latest comment ID
    discussion = await asyncio.to_thread(
        merge_request.discussions.get, previous_comment.comment_id
    )

    # Get the ID of the first note
    note_id = discussion.attributes["notes"][0]["id"]
    note = discussion.notes.get(note_id)

    # Wrap the note in <details>, indicating why.
    note.body = (
        "This comment has been superseded by a newer "
        f"Log Detective analysis.\n<details>\n{note.body}\n</details>"
    )
    await asyncio.to_thread(note.save)


async def generate_mr_comment(
    job: gitlab.v4.objects.ProjectJob,
    log_url: str,
    response: StagedResponse,
    full: bool = True,
) -> str:
    """Use a template to generate a comment string to submit to Gitlab"""

    # Locate and load the comment template
    script_path = Path(__file__).resolve().parent
    template_path = Path(script_path, "templates")
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_path))

    if full:
        tpl = jinja_env.get_template("gitlab_full_comment.md.j2")
    else:
        tpl = jinja_env.get_template("gitlab_short_comment.md.j2")

    artifacts_url = f"{job.project_url}/-/jobs/{job.id}/artifacts/download"

    if response.response_certainty >= 90:
        emoji_face = ":slight_smile:"
    elif response.response_certainty >= 70:
        emoji_face = ":neutral_face:"
    else:
        emoji_face = ":frowning2:"

    # Generate the comment from the template
    content = tpl.render(
        package=job.project_name,
        explanation=response.explanation.text,
        certainty=f"{response.response_certainty:.2f}",
        emoji_face=emoji_face,
        snippets=response.snippets,
        log_url=log_url,
        artifacts_url=artifacts_url,
    )

    return content


def _svg_figure_response(fig: matplotlib.figure.Figure):
    """Create a response with the given svg figure."""
    buf = BytesIO()
    fig.savefig(buf, format="svg", bbox_inches="tight")
    matplotlib.pyplot.close(fig)

    buf.seek(0)
    return StreamingResponse(
        buf,
        media_type="image/svg+xml",
        headers={"Content-Disposition": "inline; filename=plot.svg"},
    )


def _multiple_svg_figures_response(figures: list[matplotlib.figure.Figure]):
    """Create a response with multiple svg figures."""
    svg_contents = []
    for i, fig in enumerate(figures):
        buf = BytesIO()
        fig.savefig(buf, format="svg", bbox_inches="tight")
        matplotlib.pyplot.close(fig)
        buf.seek(0)
        svg_contents.append(buf.read().decode("utf-8"))

    html_content = "<html><body>\n"
    for i, svg in enumerate(svg_contents):
        html_content += f"<div id='figure-{i}'>\n{svg}\n</div>\n"
    html_content += "</body></html>"

    return BasicResponse(content=html_content, media_type="text/html")


class MetricRoute(str, Enum):
    """Routes for metrics"""

    ANALYZE = "analyze"
    ANALYZE_STAGED = "analyze-staged"
    ANALYZE_GITLAB_JOB = "analyze-gitlab"


class Plot(str, Enum):
    """Type of served plots"""

    REQUESTS = "requests"
    RESPONSES = "responses"
    BOTH = ""


ROUTE_TO_ENDPOINT_TYPES = {
    MetricRoute.ANALYZE: EndpointType.ANALYZE,
    MetricRoute.ANALYZE_STAGED: EndpointType.ANALYZE_STAGED,
    MetricRoute.ANALYZE_GITLAB_JOB: EndpointType.ANALYZE_GITLAB_JOB,
}


@app.get("/metrics/{route}/", response_class=StreamingResponse)
@app.get("/metrics/{route}/{plot}", response_class=StreamingResponse)
async def get_metrics(
    route: MetricRoute,
    plot: Plot = Plot.BOTH,
    period_since_now: TimePeriod = Depends(TimePeriod),
):
    """Get an handler for visualize statistics for the specified endpoint and plot."""
    endpoint_type = ROUTE_TO_ENDPOINT_TYPES[route]

    async def handler():
        """Show statistics for the specified endpoint and plot."""
        if plot == Plot.REQUESTS:
            fig = plot_engine.requests_per_time(period_since_now, endpoint_type)
            return _svg_figure_response(fig)
        if plot == Plot.RESPONSES:
            fig = plot_engine.average_time_per_responses(
                period_since_now, endpoint_type
            )
            return _svg_figure_response(fig)
        # BOTH
        fig_requests = plot_engine.requests_per_time(period_since_now, endpoint_type)
        fig_responses = plot_engine.average_time_per_responses(
            period_since_now, endpoint_type
        )
        return _multiple_svg_figures_response([fig_requests, fig_responses])

    descriptions = {
        Plot.REQUESTS: (
            "Show statistics for the requests received in the given period of time "
            f"for the /{endpoint_type.value} API endpoint."
        ),
        Plot.RESPONSES: (
            "Show statistics for responses given in the specified period of time "
            f"for the /{endpoint_type.value} API endpoint."
        ),
        Plot.BOTH: (
            "Show statistics for requests and responses in the given period of time "
            f"for the /{endpoint_type.value} API endpoint."
        ),
    }
    handler.__doc__ = descriptions[plot]

    return await handler()



import io
import logging
import zipfile

from typing import Union, Dict
from logdetective.server.models import (
    StagedResponse,
    Response,
    AnalyzedSnippet,
    Explanation,
)


LOG = logging.getLogger("logdetective")


class TextCompressor:
    """
    Encapsulates one or more texts in one or more files with the specified names
    and provides methods to retrieve them later.
    """

    def zip(self, items: Dict[str, str]) -> bytes:
        """
        Compress multiple texts into different files within a zip archive.

        Args:
            items: Dictionary where keys are file names and values are text content
                  to be compressed

        Returns:
            bytes: The compressed zip archive as bytes
        """
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
            for key, value in items.items():
                zip_file.writestr(key, value)

        zip_buffer.seek(0)
        return zip_buffer.getvalue()

    def unzip(self, zip_data: Union[bytes, io.BytesIO]) -> str:
        """
        Uncompress data created by TextCompressor.zip().

        Args:
            zip_data: A zipped stream of bytes or BytesIO object

        Returns:
            {file_name: str}: The decompressed content as a dict of file names and UTF-8 strings
        """
        if isinstance(zip_data, bytes):
            zip_buffer = io.BytesIO(zip_data)
        else:
            zip_buffer = zip_data

        content = {}
        with zipfile.ZipFile(zip_buffer, "r") as zip_file:
            file_list = zip_file.namelist()
            for file_name in file_list:
                content[file_name] = zip_file.read(file_name).decode("utf-8")

        return content


class LLMResponseCompressor:
    """
    Handles compression and decompression of LLM responses.
    """

    EXPLANATION_FILE_NAME = "explanation.txt"
    SNIPPET_FILE_NAME = "snippet_{number}.txt"
    COMPRESSOR = TextCompressor()

    def __init__(self, response: Union[StagedResponse, Response]):
        """
        Initialize with an LLM response.

        Args:
            response: Either a StagedResponse or Response object
        """
        self._response = response

    def zip_response(self) -> bytes:
        """
        Compress the content of the LLM response.

        Returns:
            bytes: Compressed response as bytes
        """
        items = {
            self.EXPLANATION_FILE_NAME: self._response.explanation.model_dump_json()
        }

        if isinstance(self._response, StagedResponse):
            for i, snippet in enumerate(self._response.snippets):
                items[self.SNIPPET_FILE_NAME.format(number=i)] = (
                    snippet.model_dump_json()
                )

        return self.COMPRESSOR.zip(items)

    @classmethod
    def unzip(
        cls, zip_data: Union[bytes, io.BytesIO]
    ) -> Union[StagedResponse, Response]:
        """
        Uncompress the zipped content of the LLM response.

        Args:
            zip_data: Compressed data as bytes or BytesIO

        Returns:
            Union[StagedResponse, Response]: The decompressed (partial) response object,
            missing response_certainty.
        """
        items = cls.COMPRESSOR.unzip(zip_data)
        if cls.EXPLANATION_FILE_NAME not in items:
            raise KeyError(
                f"Required file {cls.EXPLANATION_FILE_NAME} not found in zip archive"
            )
        explanation = Explanation.model_validate_json(items[cls.EXPLANATION_FILE_NAME])

        snippets = []
        snippet_files = {
            k: v
            for k, v in items.items()
            if cls.SNIPPET_FILE_NAME.replace("{number}.txt", "") in k
        }
        for i in range(len(snippet_files)):
            snippets.append(
                AnalyzedSnippet.model_validate_json(
                    items[cls.SNIPPET_FILE_NAME.format(number=i)]
                )
            )

        if snippets:
            response = StagedResponse(
                explanation=explanation, snippets=snippets, response_certainty=0
            )
        else:
            response = Response(explanation=explanation, response_certainty=0)

        return response



import datetime
from logging import BASIC_FORMAT
from typing import List, Dict, Optional, Literal
from pydantic import (
    BaseModel,
    Field,
    model_validator,
    field_validator,
    NonNegativeFloat,
    HttpUrl,
)
from logdetective.constants import DEFAULT_TEMPERATURE


class BuildLog(BaseModel):
    """Model of data submitted to API."""

    url: str


class JobHook(BaseModel):
    """Model of Job Hook events sent from GitLab.
    Full details of the specification are available at
    https://docs.gitlab.com/user/project/integrations/webhook_events/#job-events
    This model implements only the fields that we care about. The webhook
    sends many more fields that we will ignore."""

    # The unique job ID on this GitLab instance.
    build_id: int

    # The identifier of the job. We only care about 'build_rpm' and
    # 'build_centos_stream_rpm' jobs.
    build_name: str = Field(pattern=r"^build.*rpm$")

    # A string representing the job status. We only care about 'failed' jobs.
    build_status: str = Field(pattern=r"^failed$")

    # The kind of webhook message. We are only interested in 'build' messages
    # which represents job tasks in a pipeline.
    object_kind: str = Field(pattern=r"^build$")

    # The unique ID of the enclosing pipeline on this GitLab instance.
    pipeline_id: int

    # The unique ID of the project triggering this event
    project_id: int


class Explanation(BaseModel):
    """Model of snippet or general log explanation from Log Detective"""

    text: str
    logprobs: Optional[List[Dict]] = None

    def __str__(self):
        return self.text


class AnalyzedSnippet(BaseModel):
    """Model for snippets already processed by Log Detective.

    explanation: LLM output in form of plain text and logprobs dictionary
    text: original snippet text
    line_number: location of snippet in original log
    """

    explanation: Explanation
    text: str
    line_number: int


class Response(BaseModel):
    """Model of data returned by Log Detective API

    explanation: Explanation
    response_certainty: float
    """

    explanation: Explanation
    response_certainty: float


class StagedResponse(Response):
    """Model of data returned by Log Detective API when called when staged response
    is requested. Contains list of reponses to prompts for individual snippets.

    explanation: Explanation
    response_certainty: float
    snippets: list of AnalyzedSnippet objects
    """

    snippets: List[AnalyzedSnippet]


class InferenceConfig(BaseModel):
    """Model for inference configuration of logdetective server."""

    max_tokens: int = -1
    log_probs: int = 1
    api_endpoint: Optional[Literal["/chat/completions", "/completions"]] = (
        "/chat/completions"
    )
    url: str = ""
    api_token: str = ""
    model: str = ""
    temperature: NonNegativeFloat = DEFAULT_TEMPERATURE

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.max_tokens = data.get("max_tokens", -1)
        self.log_probs = data.get("log_probs", 1)
        self.api_endpoint = data.get("api_endpoint", "/chat/completions")
        self.url = data.get("url", "")
        self.api_token = data.get("api_token", "")
        self.model = data.get("model", "default-model")
        self.temperature = data.get("temperature", DEFAULT_TEMPERATURE)


class ExtractorConfig(BaseModel):
    """Model for extractor configuration of logdetective server."""

    context: bool = True
    max_clusters: int = 8
    verbose: bool = False

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.context = data.get("context", True)
        self.max_clusters = data.get("max_clusters", 8)
        self.verbose = data.get("verbose", False)


class GitLabConfig(BaseModel):
    """Model for GitLab configuration of logdetective server."""

    url: str = None
    api_url: str = None
    api_token: str = None

    # Maximum size of artifacts.zip in MiB. (default: 300 MiB)
    max_artifact_size: int = 300

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.url = data.get("url", "https://gitlab.com")
        self.api_url = f"{self.url}/api/v4"
        self.api_token = data.get("api_token", None)
        self.max_artifact_size = int(data.get("max_artifact_size")) * 1024 * 1024


class LogConfig(BaseModel):
    """Logging configuration"""

    name: str = "logdetective"
    level_stream: str | int = "INFO"
    level_file: str | int = "INFO"
    path: str | None = None
    format: str = BASIC_FORMAT

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.name = data.get("name", "logdetective")
        self.level_stream = data.get("level_stream", "INFO").upper()
        self.level_file = data.get("level_file", "INFO").upper()
        self.path = data.get("path")
        self.format = data.get("format", BASIC_FORMAT)


class GeneralConfig(BaseModel):
    """General config options for Log Detective"""

    packages: List[str] = None
    devmode: bool = False
    sentry_dsn: HttpUrl | None = None

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.packages = data.get("packages", [])
        self.devmode = data.get("devmode", False)
        self.sentry_dsn = data.get("sentry_dsn")


class Config(BaseModel):
    """Model for configuration of logdetective server."""

    log: LogConfig = LogConfig()
    inference: InferenceConfig = InferenceConfig()
    extractor: ExtractorConfig = ExtractorConfig()
    gitlab: GitLabConfig = GitLabConfig()
    general: GeneralConfig = GeneralConfig()

    def __init__(self, data: Optional[dict] = None):
        super().__init__()

        if data is None:
            return

        self.log = LogConfig(data.get("log"))
        self.inference = InferenceConfig(data.get("inference"))
        self.extractor = ExtractorConfig(data.get("extractor"))
        self.gitlab = GitLabConfig(data.get("gitlab"))
        self.general = GeneralConfig(data.get("general"))


class TimePeriod(BaseModel):
    """Specification for a period of time.

    If no indication is given
    it falls back to a 2 days period of time.

    Can't be smaller than a hour"""

    weeks: Optional[int] = None
    days: Optional[int] = None
    hours: Optional[int] = None

    @model_validator(mode="before")
    @classmethod
    def check_exclusive_fields(cls, data):
        """ Check that only one key between weeks, days and hours is defined"""
        if isinstance(data, dict):
            how_many_fields = sum(
                1
                for field in ["weeks", "days", "hours"]
                if field in data and data[field] is not None
            )

            if how_many_fields == 0:
                data["days"] = 2  # by default fallback to a 2 days period

            if how_many_fields > 1:
                raise ValueError("Only one of months, weeks, days, or hours can be set")

        return data

    @field_validator("weeks", "days", "hours")
    @classmethod
    def check_positive(cls, v):
        """Check that the given value is positive"""
        if v is not None and v <= 0:
            raise ValueError("Time period must be positive")
        return v

    def get_time_period(self) -> datetime.timedelta:
        """Get the period of time represented by this input model.

        Returns:
            datetime.timedelta: The time period as a timedelta object.
        """
        delta = None
        if self.weeks:
            delta = datetime.timedelta(weeks=self.weeks)
        elif self.days:
            delta = datetime.timedelta(days=self.days)
        elif self.hours:
            delta = datetime.timedelta(hours=self.hours)
        return delta

    def get_period_start_time(
        self, end_time: datetime.datetime = None
    ) -> datetime.datetime:
        """Calculate the start time of this period based on the end time.

        Args:
            end_time (datetime.datetime, optional): The end time of the period.
                Defaults to current UTC time if not provided.

        Returns:
            datetime.datetime: The start time of the period.
        """
        time = end_time or datetime.datetime.now(datetime.timezone.utc)
        if end_time.tzinfo is None:
            end_time = end_time.replace(tzinfo=datetime.timezone.utc)
        return time - self.get_time_period()



import logging
import yaml
from logdetective.constants import SNIPPET_DELIMITER
from logdetective.server.models import Config, AnalyzedSnippet


def format_analyzed_snippets(snippets: list[AnalyzedSnippet]) -> str:
    """Format snippets for submission into staged prompt."""
    summary = f"\n{SNIPPET_DELIMITER}\n".join(
        [
            f"[{e.text}] at line [{e.line_number}]: [{e.explanation.text}]"
            for e in snippets
        ]
    )
    return summary


def load_server_config(path: str | None) -> Config:
    """Load configuration file for logdetective server.
    If no path was provided, or if the file doesn't exist, return defaults.
    """
    if path is not None:
        try:
            with open(path, "r") as config_file:
                return Config(yaml.safe_load(config_file))
        except FileNotFoundError:
            # This is not an error, we will fall back to default
            print("Unable to find server config file, using default then.")
    return Config()


def get_log(config: Config):
    """
    Initialize a logger for this server
    """
    log = logging.getLogger(config.log.name)
    if getattr(log, "initialized", False):
        return log

    log.setLevel("DEBUG")

    # Drop the default handler, we will create it ourselves
    log.handlers = []

    # STDOUT
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(config.log.format))
    stream_handler.setLevel(config.log.level_stream)
    log.addHandler(stream_handler)

    # Log to file
    if config.log.path:
        file_handler = logging.FileHandler(config.log.path)
        file_handler.setFormatter(logging.Formatter(config.log.format))
        file_handler.setLevel(config.log.level_file)
        log.addHandler(file_handler)

    log.initialized = True
    return log
